# -*- coding: utf-8 -*-
"""
/***************************************************************************
 TPExporter
                                 A QGIS plugin
 Модуль для экспорта векторных слоев из состава проектов документов территориального планирования в формат GML для последующего размещения на официальном сайте Федеральной государственной информационной системы территориального планирования
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2024-05-19
        git sha              : $Format:%H$
        copyright            : (C) 2024 by Иван Гевало
        email                : volen93@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QSettings, QTranslator, QCoreApplication
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QTreeWidgetItemIterator, QTableWidgetItem
from PyQt5.QtXml import QDomDocument

# Initialize Qt resources from file resources.py
from .resources import *
# Import the code for the dialog
from .tp_exporter_dialog import TPExporterDialog

from qgis.core import QgsProject, QgsMapLayer, QgsCoordinateReferenceSystem, QgsCoordinateTransform
from osgeo import ogr
from qgis import processing

import os.path

from datetime import datetime

class TPExporter:
    """QGIS Plugin Implementation."""

    MAIN_BUTTON_EXEC_FUNCTION = 1
    MAIN_BUTTON_CLOSE_FUNCTION = 2

    prjLayers = []

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'TPExporter_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)
            QCoreApplication.installTranslator(self.translator)

        # Declare instance attributes
        self.actions = []
        self.menu = self.tr(u'&Terplan Exporter')

        # Check if plugin was started the first time in current QGIS session
        # Must be set in initGui() to survive plugin reloads
        self.first_start = None

        self.mainButtonFumctionFlag = self.MAIN_BUTTON_EXEC_FUNCTION


    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('TPExporter', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            # Adds plugin icon to Plugins toolbar
            self.iface.addToolBarIcon(action)

        if add_to_menu:
            self.iface.addPluginToVectorMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/tp_exporter/icon.png'
        self.add_action(
            icon_path,
            text=self.tr(u'Осуществить экспорт в GML'),
            callback=self.run,
            parent=self.iface.mainWindow())

        # will be set False in run()
        self.first_start = True


    def unload(self):
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginVectorMenu(
                self.tr(u'&Terplan Exporter'),
                action)
            self.iface.removeToolBarIcon(action)


    def filterClasses(self, geomType):
        itemIterator = QTreeWidgetItemIterator(self.dlg.classesTree, QTreeWidgetItemIterator.NoChildren)

        while itemIterator.value() is not None:
            if itemIterator.value().toolTip(0).find(geomType) != -1:
                itemIterator.value().setDisabled(False)
            else:
                itemIterator.value().setDisabled(True)

            itemIterator.__iadd__(1)


    def getLayerById(self, layerId):
        for cLayer in self.prjLayers:
            if cLayer.id() == layerId:
                return cLayer
        

    def getLayerFromComboBox(self):
        if len(self.prjLayers) > 0:
            return self.prjLayers[self.dlg.projectLayersComboBox.currentIndex()]
        
        return None


    def onProjectLayersCBoxChanged(self):
        cLayer = self.getLayerFromComboBox()

        if cLayer == None:
            return
        
        cLayerGeomType = cLayer.wkbType()

        if cLayerGeomType in [2, 5, 11]:
            self.filterClasses("линейный")
        elif cLayerGeomType in [1, 4]:
            self.filterClasses("точечный")
        elif cLayerGeomType in [3, 6]:
            self.filterClasses("площадной")


    def updateMainButtonState(self):
        if self.dlg.layersTable.rowCount() > 0 and len(self.dlg.xsdName.text()) > 0 and self.dlg.fileSelector.filePath() != "":
            self.dlg.mainButton.setEnabled(True)
        elif self.mainButtonFumctionFlag == self.MAIN_BUTTON_EXEC_FUNCTION:
            self.dlg.mainButton.setEnabled(False)


    # Needs rewriting
    def transformCRS(self, layerCRS):
        destCRS = QgsCoordinateReferenceSystem("EPSG:3857")

        if layerCRS.authid() == "EPSG:3857":
            return None
        else:
            transform = QgsCoordinateTransform(layerCRS, destCRS, QgsProject.instance())
            return transform


    def transformLayerToMultiType(self, layer):
        params = {'INPUT' : layer.source(), 'OUTPUT' : 'TEMPORARY_OUTPUT'}

        return processing.run("native:promotetomulti", params)


    def buildGMLTree(self):
        gmlOptions = ["FORMAT=GML32"]
        geomDocument = QDomDocument()

        cRowIndex = 0
        layerCRS = self.getLayerById(self.dlg.layersTable.item(cRowIndex, 0).toolTip()).crs()

        self.dlg.log.append("<strong>\nПостроение GML-дерева:</strong>")
        gmlDocument = QDomDocument()
        self.dlg.log.append("- формирование корневого элемента")
        root = gmlDocument.createElement("gml:FeatureCollection")
        root.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
        root.setAttribute("xmlns", "http://fgistp")
        root.setAttribute("xmlns:gml", "http://www.opengis.net/gml/3.2")
        root.setAttribute("xsi:schemaLocation", f"http://fgistp {self.dlg.xsdName.text()}")
        root.setAttribute("gml:id", f"i{hash(datetime.now().__str__)}")
        self.dlg.log.append("- добавление корневого элемента в дерево")
        gmlDocument.appendChild(root)

        totalObjectsCount = 0

        while True:
            if self.dlg.layersTable.item(cRowIndex, 0) is None:
                break

            cLayer = self.getLayerById(self.dlg.layersTable.item(cRowIndex, 0).toolTip())
            featureCodeName = self.dlg.layersTable.item(cRowIndex, 1).text()
            #features = self.getLayerById(self.dlg.layersTable.item(cRowIndex, 0).toolTip()).getFeatures()

            objectsCount = 0

            self.dlg.log.append(f"Обработка слоя <strong>'{self.dlg.layersTable.item(cRowIndex, 0).text()}'</strong>")

            self.dlg.log.append("- преобразование в составные типы")
            features = self.transformLayerToMultiType(cLayer)['OUTPUT'].getFeatures()

            for feature in features:
                if feature.hasGeometry():
                    featureMember = gmlDocument.createElement("gml:featureMember")
                    root.appendChild(featureMember)
                    cFeature = gmlDocument.createElement(featureCodeName)
                    cFeature.setAttribute("gml:id", f"{featureCodeName}.{objectsCount}")
                    attributeSchema = feature.__geo_interface__["properties"]

                    # добавление атрибутов объекта в GML дерево
                    for key, value in attributeSchema.items():
                        addFlag = False
                        valueType = type(value)

                        if str(valueType) == "<class 'int'>":
                            if int(value) > 0 and str(key) != "PASS_TERM" and str(key) != "SUBURBAN_TR":
                                addFlag = True
                        elif str(valueType) == "<class 'float'>":
                            if float(value) > 0.0:
                                addFlag = True
                        elif str(valueType) == "<class 'str'>":
                            if len(str(value)) > 0 and str(value) != "NULL":
                                addFlag = True

                        if addFlag == True:
                            attr = gmlDocument.createElement(f"{str(key).upper()}")
                            attrValue = gmlDocument.createTextNode(f"{value}")
                            attr.appendChild(attrValue)
                            cFeature.appendChild(attr)

                    # запись геометрии
                    ogrGeom = ogr.CreateGeometryFromWkt(feature.geometry().asWkt())
                    geomDocument.setContent(ogrGeom.ExportToGML(gmlOptions))
                    geom = geomDocument.firstChild().toElement()
                    geom.setAttribute("srsName", layerCRS.authid())

                    cFeature.appendChild(geom)

                    featureMember.appendChild(cFeature)
                    objectsCount += 1

            self.dlg.log.append(f"- количество добавленных объектов: {objectsCount}")
            
            cRowIndex += 1
            totalObjectsCount += objectsCount

        self.dlg.log.append(f"<strong>Общее количество добавленных объектов: {totalObjectsCount}</strong>")
        self.dlg.log.append("<strong>\n\nЗапись в файл...</strong>")

        return gmlDocument.toString(4)


    def saveAsGML(self):
        try:
            gmlFile = open(self.dlg.fileSelector.filePath(), "w")

            try:
                gmlFile.write(self.buildGMLTree())
                self.dlg.log.append(f"<strong style='color:green'>Файл {self.dlg.fileSelector.filePath()} успешно создан.</strong>")
            finally:
                gmlFile.close()
        except Exception as ex:
            self.dlg.log.append(f"<strong style='color:red'>{ex.__str__()}</strong>")


    def onMainButtonClicked(self):
        if self.mainButtonFumctionFlag == self.MAIN_BUTTON_EXEC_FUNCTION:
            self.dlg.tabs.setCurrentIndex(2)

            self.dlg.log.append("<strong>Начало работы модуля</strong>")

            self.saveAsGML()
            self.dlg.cancelButton.hide()
            self.dlg.mainButton.setText("Закрыть")
            self.mainButtonFumctionFlag = self.MAIN_BUTTON_CLOSE_FUNCTION

            if self.dlg.clearTableCheckBox.isChecked():
                self.clearLayersTable()

        elif self.mainButtonFumctionFlag == self.MAIN_BUTTON_CLOSE_FUNCTION:
            self.dlg.hide()


    def onCancelButtonClicked(self):
        self.dlg.hide()


    def updateAddLayerButtonState(self):
        selectedClass = self.dlg.classesTree.currentItem()

        if self.dlg.projectLayersComboBox.currentText() != "" and (selectedClass is not None and selectedClass.childCount() == 0):
            self.dlg.addLayerButton.setEnabled(True)
        else:
            self.dlg.addLayerButton.setEnabled(False)


    def onAddLayerButtonClicked(self):
        cRow = self.dlg.layersTable.rowCount()
        cLayer = self.getLayerFromComboBox()

        self.dlg.layersTable.insertRow(cRow)

        layerDesc = QTableWidgetItem(cLayer.name())
        layerDesc.setToolTip(cLayer.id())

        self.dlg.layersTable.setItem(cRow, 0, layerDesc)
        self.dlg.layersTable.setItem(cRow, 1, QTableWidgetItem(self.dlg.classesTree.currentItem().text(1)))

        if not self.dlg.clearTableButton.isEnabled():
            self.dlg.clearTableButton.setEnabled(True)


    def onDeleteLayerButtonClicked(self):
         if self.dlg.layersTable.rowCount() == 1:
             self.clearLayersTable()
             return

         if self.dlg.layersTable.rowCount() > 0:
            selectedRowIndex = self.dlg.layersTable.selectedIndexes()[0].row()
            
            if selectedRowIndex is not None:
                self.dlg.layersTable.removeRow(selectedRowIndex)

            self.updateMainButtonState()


    def clearLayersTable(self):
        while True:
            if self.dlg.layersTable.item(0, 0) is None:
                break

            self.dlg.layersTable.removeRow(0)

        self.dlg.clearTableButton.setEnabled(False)
        self.updateMainButtonState()


    def initPrjLayersList(self):
        projectLayers = list(QgsProject.instance().mapLayers().values())

        if len(projectLayers) > 0:
            for cLayer in projectLayers:
                if cLayer.type() == QgsMapLayer.VectorLayer and cLayer.isSpatial():
                    self.prjLayers.append(cLayer)
                    self.dlg.projectLayersComboBox.addItem(cLayer.name())


    def resetFormState(self):
        self.dlg.projectLayersComboBox.clear()
        self.dlg.tabs.setCurrentIndex(0)
        self.dlg.log.clear()
        self.dlg.cancelButton.show()
        self.updateAddLayerButtonState()
        self.clearLayersTable()

        self.mainButtonFumctionFlag = self.MAIN_BUTTON_EXEC_FUNCTION
        self.dlg.mainButton.setText("Выполнить")
        self.dlg.mainButton.setEnabled(False)

        self.dlg.clearTableButton.setEnabled(False)
        self.dlg.fileSelector.setFilePath("")
        self.dlg.classesTree.reset()


    def run(self):
        """Run method that performs all the real work"""

        # Create the dialog with elements (after translation) and keep reference
        # Only create GUI ONCE in callback, so that it will only load when the plugin is started
        if self.first_start == True:
            self.first_start = False
            self.dlg = TPExporterDialog()

            self.dlg.projectLayersComboBox.currentTextChanged.connect(self.onProjectLayersCBoxChanged)
            self.dlg.classesTree.itemSelectionChanged.connect(self.updateAddLayerButtonState)
            self.dlg.xsdName.textChanged.connect(self.updateMainButtonState)
            self.dlg.fileSelector.fileChanged.connect(self.updateMainButtonState)
            self.dlg.cancelButton.clicked.connect(self.onCancelButtonClicked)
            self.dlg.mainButton.clicked.connect(self.onMainButtonClicked)
            self.dlg.addLayerButton.clicked.connect(self.onAddLayerButtonClicked)
            self.dlg.deleteLayerButton.clicked.connect(self.onDeleteLayerButtonClicked)
            self.dlg.clearTableButton.clicked.connect(self.clearLayersTable)

        self.resetFormState()
        self.initPrjLayersList()
        self.onProjectLayersCBoxChanged()

        # show the dialog
        self.dlg.show()

        # Run the dialog event loop
        result = self.dlg.exec_()
        # See if OK was pressed
        if result:
            # Do something useful here - delete the line containing pass and
            # substitute with your code.
            pass
